struct Uniforms {
	val_seq_1:         vec4<i32>,
	val_seq_2:         vec4<i32>,
	val_seq_env:       vec2<f32>,
	shader_idx:        u32,
	elapsed_time:      f32,
	scene_time:        f32,
	frame_time:        f32,
	sequencer_ticks:   u32,
	screen_flash:      f32,
	screen_width:      i32,
	screen_height:     i32,
	track_volume:      f32,

	pad1: u32,

	val_mod0:          vec4<f32>,
	val_mod1:          vec4<f32>,
	val_mod2:          vec4<f32>,
	val_mod3:          vec4<f32>,
};

@group(0) @binding(0) var sampler_bilinear: sampler;

@group(0) @binding(1) var tex_charset:      texture_2d<f32>;
@group(0) @binding(2) var tex_colormap:     texture_2d<f32>;
@group(0) @binding(3) var tex_text_buffer:  texture_2d<u32>;

@group(0) @binding(4) var<uniform> uniforms: Uniforms;

@group(0) @binding(5) var tex_bg_image_0:   texture_2d<f32>;
@group(0) @binding(6) var tex_bg_image_1:   texture_2d<f32>;
@group(0) @binding(7) var tex_bg_image_2:   texture_2d<f32>;
@group(0) @binding(8) var tex_bg_image_3:   texture_2d<f32>;
@group(0) @binding(9) var tex_vis_fft:      texture_2d<f32>;


/* this is idiotic. should rewrite this properly */
@group(1) @binding(0) var tex_ping:         texture_2d<f32>;
@group(1) @binding(1) var tex_ping2:        texture_2d<f32>;

@group(1) @binding(2) var tex_storage_pong: texture_storage_2d<rgba16float,write>;


const filter_weights = array(0.375, 0.25, 0.0625);

struct VSOut
{
	@builtin(position) position: vec4f,
	@location(0) uv: vec2f,
};

struct VSOutLine
{
	@builtin(position) position: vec4f,
	@location(0) uv: vec2f,

	@location(1) @interpolate(flat) p1: vec2f,
	@location(2) @interpolate(flat) p2: vec2f,

	@location(3) @interpolate(flat) width: f32,

	@location(4) pos: vec2f,
};

@vertex
fn vs_main(@builtin(vertex_index) in_vertex_index: u32) -> VSOut
{
	var out: VSOut;

	if(in_vertex_index == 0u) {
		out.position = vec4f(-1.0, -3.0, 0.0, 1.0);
	} else if(in_vertex_index == 1u) {
		out.position = vec4f(-1.0, 1.0, 0.0, 1.0);
	} else {
		out.position = vec4f(3.0, 1.0, 0.0, 1.0);
	}

	out.uv = out.position.xy * 0.5 + vec2(0.5);
	return out;
}

fn get_vertex(idx: i32) -> vec2f
{
	if(idx <= 0) {
		return vec2f(-0.5 + sin(uniforms.scene_time * 1.2345) * 0.2, -0.5);
	}
	if(idx == 1) {
		return vec2f(-0.5, 0.0 + sin(uniforms.scene_time * 2.345) * 0.1);
	}
	if(idx == 2) {
		return vec2f(0.5, 0.0);
	}
	if(idx >= 3) {
		return vec2f(0.5, 0.0) + 1.0 * vec2f(sin(uniforms.scene_time), cos(uniforms.scene_time));
	}
	return vec2f(0.0);
}

fn get_dir(idx: i32, num_line_segments: i32) -> vec2f
{
	let curr = min(num_line_segments - 1, max(idx, 0));
	let next = min(num_line_segments,     max(idx, 0) + 1);
	return normalize(get_vertex(next) - get_vertex(curr));
}

fn rotate(x: vec2f) -> vec2f
{
	return vec2f(-x.y, x.x);
}

fn sqr_dist_point_to_line_segment(p1: vec2f, p2: vec2f, x: vec2f) -> f32
{
    let  x_p1 =  x - p1;
    let  x_p2 =  x - p2;
    let p2_p1 = p2 - p1;

    let dot_p2_p1 = dot(p2_p1, p2_p1);
    let prj       = max(0.0, dot((x_p1), (p2_p1)));

	/*
    let dd = prj < dot_p2_p1
        ? dot(x_p1, x_p1) - prj * prj / dot_p2_p1
        : dot(x_p2, x_p2);
		*/

	let dd = select(
		dot(x_p2, x_p2),
		dot(x_p1, x_p1) - prj * prj / dot_p2_p1,
		prj < dot_p2_p1);

    return max(0.0, dd);
}

@vertex
fn vs_line(@builtin(vertex_index) in_vertex_index: u32) -> VSOutLine
{
	var out: VSOutLine;

	let line_idx = i32(in_vertex_index) / 6;
	let vert_idx = i32(in_vertex_index) % 6;

	let num_line_segments = 3;

	let p_prev = get_vertex(line_idx - 1);
	let p_curr = get_vertex(line_idx);
	let p_next = get_vertex(line_idx + 1);

	let dir_curr = get_dir(line_idx,     num_line_segments);
	let dir_prev = get_dir(line_idx - 1, num_line_segments);
	let dir_next = get_dir(line_idx + 1, num_line_segments);

	let tangent_prev = rotate(normalize(dir_curr + dir_prev));
	let tangent_next = rotate(normalize(dir_curr + dir_next));

	let normal = vec2f(-dir_curr.y, dir_curr.x);

	out.width = 0.01 + 0.005 * sin(uniforms.scene_time);
	let thickness = out.width + 4.0 / f32(uniforms.screen_height);

	var joint_prev: vec2f;
	var joint_next: vec2f;

	var quad = array<vec2f, 4>();
	if(false) {
		joint_prev = tangent_prev * thickness / max(0.3, (dot(tangent_prev, normal)));
		joint_next = tangent_next * thickness / max(0.3, (dot(tangent_next, normal)));
		quad[0] = p_curr - joint_prev;
		quad[1] = p_curr + joint_prev;
		quad[2] = p_next - joint_next; // - mix(foo, joint_prev, f);
		quad[3] = p_next + joint_next;
	}
	else {
		joint_prev = thickness * rotate(dir_curr);
		joint_next = joint_prev;
		quad[0] = p_curr - joint_prev - dir_curr * thickness;
		quad[1] = p_curr + joint_prev - dir_curr * thickness;
		quad[2] = p_next - joint_next + dir_curr * thickness; // - mix(foo, joint_prev, f);
		quad[3] = p_next + joint_next + dir_curr * thickness;
	}

	let f = 1.0;

	//quad[3] = p_next + mix(foo, joint_prev, f);

	let len = distance(p_curr, p_next);

	var uv_arr = array<vec2f, 4>();
	uv_arr[0] = vec2f(0.0, 0.0);
	uv_arr[1] = vec2f(0.0, 1.0);
	uv_arr[2] = vec2f(len, 0.0);
	uv_arr[3] = vec2f(len, 1.0);

	var p = vec2f(0.0);
	var uv = vec2f(0.0);
	switch(vert_idx) {
	case 0: { p = quad[0]; uv = uv_arr[0]; }
	case 1: { p = quad[2]; uv = uv_arr[2]; }
	case 2: { p = quad[1]; uv = uv_arr[1]; }
	case 3: { p = quad[1]; uv = uv_arr[1]; }
	case 4: { p = quad[2]; uv = uv_arr[2]; }
	case 5: { p = quad[3]; uv = uv_arr[3]; }
	default: {}
	}

	let a = f32(uniforms.screen_height) / f32(uniforms.screen_width);

	out.position = vec4f(p.x * a, p.y, 0.0, 1.0);

	out.uv = uv;

	out.p1 = p_curr;
	out.p2 = p_next;

	out.pos = p;

	return out;
}

@fragment
fn fs_line(in: VSOutLine) -> @location(0) vec4f
{
	let r = in.width;

	var coverage       = 0.0;
	var coverage_inner = 1.0;

	if(false) {
		let dp_dx = dpdxFine(in.pos);
		let dp_dy = dpdyFine(in.pos);

		let filter_radius = 2;
		for(var yy = -filter_radius; yy <= filter_radius; yy++) {
			for(var xx = -filter_radius; xx <= filter_radius; xx++) {
				let delta = 0.5 * vec2f(f32(xx) / f32(filter_radius), f32(yy) / f32(filter_radius));
				let p = in.pos + delta.x * dp_dx + delta.y * dp_dy;
				let d = sqr_dist_point_to_line_segment(in.p1, in.p2, p);
				coverage += 1.0 - step(r * r, d);
			}
		}

		coverage /= f32((filter_radius * 2 + 1) * (filter_radius * 2 + 1));
	}
	else {
		let d = sqrt(sqr_dist_point_to_line_segment(in.p1, in.p2, in.pos));
		let fwidth_d = fwidth(d);
		//coverage = 1.0 - step(r * r, d);
		coverage = 1.0 - smoothstep(r, r + fwidth_d, d);
	}

	{
		let d = sqrt(sqr_dist_point_to_line_segment(in.p1, in.p2, in.pos));
		let fwidth_d = fwidth(d);
		let r_inner = r * 0.8;
		coverage_inner = 1.0 - smoothstep(r_inner, r_inner + fwidth_d, d);
	}

	{
		let d = sin(in.uv.x * 32.0 + uniforms.scene_time * 16.0) * 0.5 + 0.5;
		//let d = fract((in.uv.x * 32.0 - uniforms.scene_time * 2.0));
		let fwidth_d = fwidth(d);
		coverage_inner *= 1.0 - smoothstep(0.2, 0.2 + fwidth_d, d);
	}
	//let s = fract((in.uv.x * 8.0 - uniforms.scene_time * 2.0)) < 0.05;
	//let s = sin(in.uv.x * 3.141 * 2.0 * 20.0 + uniforms.scene_time * 10.0) * 0.5 + 0.5;

	let c = vec3f(0.7);

	//return vec4f(coverage, coverage, coverage, 1.0);
	return vec4f(c, coverage * (1.0 - coverage_inner));
	//return vec4f(in.uv.x, in.uv.y, 0.0, 1.0);
}

fn get_offset_uv(uv: vec2f, x: i32, y: i32, r: f32) -> vec4f
{
	var uv_off = uv;
	uv_off.x += f32(x) * r * dpdxFine(uv.x);
	uv_off.y += f32(y) * r * dpdxFine(uv.y);

	return textureSample(tex_ping,  sampler_bilinear, uv_off);
}

// Catmull-Rom filtering code from http://vec3.ca/bicubic-filtering-in-fewer-taps/
fn read_img_catmull_rom(_p: vec2f) -> vec4f
{
	let tex_size = textureDimensions(tex_ping);
	let mul = vec2f(tex_size.xy);
	let div = 1.0 / mul;

	let p = _p * mul;
    let tc = floor(p - 0.5) + 0.5;
    let f = p - tc;
    let f2 = f * f;
    let f3 = f2 * f;

    let w0 = f2 - 0.5 * (f3 + f);
    let w1 = 1.5 * f3 - 2.5 * f2 + 1.0;
    let w3 = 0.5 * (f3 - f2);
    let w2 = 1.0 - w0 - w1 - w3;

    let w12 = w1 + w2;

    let tc0 = tc - 1.0;
    let tc12 = tc + w2 / w12;
    let tc3 = tc + 2.0;

    let result =
        textureSample(tex_ping, sampler_bilinear, vec2(tc0.x,  tc0.y ) * div) * (w0.x  * w0.y ) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc0.x,  tc12.y) * div) * (w0.x  * w12.y) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc0.x,  tc3.y ) * div) * (w0.x  * w3.y ) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc12.x, tc0.y ) * div) * (w12.x * w0.y ) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc12.x, tc12.y) * div) * (w12.x * w12.y) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc12.x, tc3.y ) * div) * (w12.x * w3.y ) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc3.x,  tc0.y ) * div) * (w3.x  * w0.y ) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc3.x,  tc12.y) * div) * (w3.x  * w12.y) +
        textureSample(tex_ping, sampler_bilinear, vec2(tc3.x,  tc3.y ) * div) * (w3.x  * w3.y );

    return result;
}


@fragment
fn fs_main(in: VSOut) -> @location(0) vec4f
{
	let uv = vec2(in.uv.x, 1.0 - in.uv.y);
	var color_blurred = textureSample(tex_ping2, sampler_bilinear, uv).rgb;

	color_blurred = pow(color_blurred, vec3(1.3, 1.3, 1.3));

	var color = read_img_catmull_rom(uv);
	color = max(vec4(0.0), color);

	if(color.a > 1.0) {
		color_blurred *= 1.1 + 0.3 * sin(uniforms.elapsed_time * 3.14159 * 2.0);
	}

	var c = color.rgb + color_blurred * 0.45;
	c *= (0.9 + 0.1 * sin(in.position.y * 3.14159));

	if false && abs(uv.x - 0.5) < 2.0 * dpdxFine(uv.x) {
		c += vec3f(1.0);
	}

	return vec4f(c, 1.0);
}

@fragment
fn fs_screen_saver(in: VSOut) -> @location(0) vec4f
{
	let uv = vec2(in.uv.x, 1.0 - in.uv.y);

	let aspect = f32(uniforms.screen_width) / f32(uniforms.screen_height);
	let uv_fx = uv * vec2(aspect, 1.0);
	let bg_effect = noise_sum_abs_sin(vec3f(uv_fx, 1000.0 * sin(uniforms.elapsed_time * 0.00001)));

	return vec4f(sin(bg_effect * 40.0 + vec3f(0.1111, 0.2345, 0.314314) * uniforms.elapsed_time * 1.0) * 0.25 + 0.25, 1.0);
}

/*

//original shader by Doctor Birdy

#define t iTime
#define res iResolution.xy

void mainImage(out vec4 fragColor, in vec2 fragCoord) {
    vec3 col = vec3(0.0);
    float lum;
    vec2 mouse = iMouse.xy / res; // Normalized mouse position
    float zoom = t + mouse.x * 10.0; // Use mouse X for zoom speed
    float rotation = mouse.y * 6.28; // Use mouse Y for rotation

        vec2 _uv = fragCoord / res;
        _uv = _uv * 2.0 - 1.0;
        _uv.x *= res.x / res.y;
        float s = sin(rotation), c = cos(rotation);
        _uv = vec2(_uv.x * c - _uv.y * s, _uv.x * s + _uv.y * c);
    for (int i = 0; i < 3; i++) {
        vec2 uv = _uv;
        
        vec2 radialOffset = vec2(sin(zoom + length(uv) * 8.0), cos(zoom + length(uv) * 8.0));
        uv += radialOffset * 0.2;
        lum = 0.2 / (length(mod(uv * 1.0, 1.0) - 0.5) + 0.01);
        
        // Modify this line to introduce color shifting
        col[i] = lum * abs(sin(zoom + length(uv) * 1.0 + float(i) * 0.3));
    }
    
    // Enhance overall saturation
    col = pow(col, vec3(1.3));
    
    fragColor = vec4(col, 1.0);
}

*/

fn mymod(x: vec2f, y: vec2f) -> vec2f
{
	return fract(x / y) * y;
}

@fragment
fn fs_background(in: VSOut) -> @location(0) vec4f
{
	let uv = vec2(in.uv.x, 1.0 - in.uv.y) * 2.0 - 1.0;

	let aspect = f32(uniforms.screen_width) / f32(uniforms.screen_height);
	let uv_fx = uv * vec2(aspect, 1.0);

	var bg_color = vec4f(0.0, 0.0, 0.0, 1.0);

	switch uniforms.shader_idx {
		case 1u: {
			bg_color = vec4f(render_sine(uv_fx) * 0.1, 1.0);
		}
		case 2u: {
			bg_color = vec4f(render_distance_field(uv_fx), 1.0);
		}
		case 3u: {
			let tex_size = textureDimensions(tex_bg_image_0);
			let a = f32(tex_size.y) / f32(tex_size.x) * aspect;
			let uv_bg = vec2f(in.uv.x * a + (1.0 - a) / 2.0, 1.0 - in.uv.y);
			var c = textureSample(tex_bg_image_0, sampler_bilinear, uv_bg);
			if(uv_bg.x < 0.0 || uv_bg.x > 1.0) {
				c = vec4f(0.0, 0.0, 0.0, 1.0);
			}
			bg_color = c;
		}
		case 4u: {
			let tex_size = textureDimensions(tex_bg_image_1);
			let a = f32(tex_size.y) / f32(tex_size.x) * aspect;
			let uv_bg = vec2f(in.uv.x * a + (1.0 - a) / 2.0, 1.0 - in.uv.y);
			var c = textureSample(tex_bg_image_1, sampler_bilinear, uv_bg);
			if(uv_bg.x < 0.0 || uv_bg.x > 1.0) {
				c = vec4f(0.0, 0.0, 0.0, 1.0);
			}
			bg_color = c;
		}
		case 5u: {
			let tex_size = textureDimensions(tex_bg_image_2);
			let a = f32(tex_size.y) / f32(tex_size.x) * aspect;
			let uv_bg = vec2f(in.uv.x * a + (1.0 - a) / 2.0, 1.0 - in.uv.y);
			var c = textureSample(tex_bg_image_2, sampler_bilinear, uv_bg);
			if(uv_bg.x < 0.0 || uv_bg.x > 1.0) {
				c = vec4f(0.0, 0.0, 0.0, 1.0);
			}
			bg_color = c;
		}
		case 6u: {
			let tex_size = textureDimensions(tex_bg_image_3);
			let a = f32(tex_size.y) / f32(tex_size.x) * aspect;
			let uv_bg = vec2f(in.uv.x * a + (1.0 - a) / 2.0, 1.0 - in.uv.y);
			var c = textureSample(tex_bg_image_3, sampler_bilinear, uv_bg);
			if(uv_bg.x < 0.0 || uv_bg.x > 1.0) {
				c = vec4f(0.0, 0.0, 0.0, 1.0);
			}
			bg_color = c;
		}
		case 7u: {
			let uv_scaled = uv_fx * 1.0 * f32(uniforms.val_seq_1.y + 1); 
			let color_fg = textureLoad(tex_colormap, vec2i(1, 0), 0).rgb;

			var c = vec3f(0);

			for(var i = 0; i < min(8, uniforms.val_seq_1.w + 1); i++) {
				let uv_oct = uv_scaled * pow(2.0, f32(i));
				let p = vec2i(uv_oct) + i;
				let v = (p.x ^ p.y ^ uniforms.val_seq_1.z) % (i32(uniforms.val_seq_1.x + 2));

				let char = textureSample(tex_ping2, sampler_bilinear, fract(uv_oct)).rgb;
				if(v == 0) {
					c = mix(char, c, 0.2) + 0.05 * color_fg;
				}
				else {
					c = mix(char, c, 1.0);
				}
			}

			bg_color = vec4f(c * 0.4, 1.0);
		}
		case 8u: {
			let x = in.uv.x * 16.0 - 0.25;
			let y = in.uv.y;

			bg_color = vec4f(x, y, 0.0, 1.0);

			let idx = i32(x);

			var v = 0.0;
			if(idx < 4) {
				v = uniforms.val_mod0[idx % 4];
			}
			else if(idx < 8) {
				v = uniforms.val_mod1[idx % 4];
			}
			else if(idx < 12) {
				v = uniforms.val_mod2[idx % 4];
			}
			else if(idx < 16) {
				v = uniforms.val_mod3[idx % 4];
			}

			var color_fg = vec3f(1, 0, 1);

			if(i32(idx) % 2 > 0) {
				color_fg = textureLoad(tex_colormap, vec2i(14, 0), 0).rgb;
			}

			let bar = sin(fract(x) * 3.14159 * 2.0) * 0.5 + 0.5;
			if(v > y && fract(x) < 0.5) {
				bg_color = vec4f(color_fg * bar * 0.25, 1.0);
			}
			else {
				bg_color = vec4f(0.0, 0.0, 0.0, 1.0);
			}
		}
		case 9u: {
			bg_color = vec4f(render_tunnel(uv_fx) * 0.1, 1.0);
		}
		case 10u: {
			bg_color = vec4f(render_xor_circles(uv_fx), 1.0);
		}
		case 11u: {
			let color1 = textureLoad(tex_colormap, vec2i(14, 0), 0);
			let color2 = textureLoad(tex_colormap, vec2i(15, 0), 0);

			let num_aa_samples = 8;
			var cvg = 0.0;
			let aa_offset = dpdxFine(in.uv.x) * (1.0 / f32(num_aa_samples));
			let cmp = select(0.5 - in.uv.y, in.uv.y - 0.5, in.uv.y > 0.5) * 2.0;
			var stereo = 0.0;
			for(var i = 0; i < num_aa_samples; i++) {
				let freq = (pow(20.0, in.uv.x + f32(i) * aa_offset) - 1.0) / 19.0;
				let l_r = textureSample(tex_vis_fft, sampler_bilinear, vec2f(freq, 0.0)).rg;

				let mi = min(l_r.x, l_r.y);
				let ma = max(l_r.x, l_r.y);
				let spectrum = select(l_r.x, l_r.y, in.uv.y > 0.5);

				cvg    += f32(spectrum >= cmp) * clamp(cmp / max(mi, 1e-4), 0.0, 1.0);
				stereo += pow(clamp((cmp - mi) / max(ma - mi, 1e-4), 0.0, 1.0), 4.0);
			}
			cvg    *= 1.0 / f32(num_aa_samples);
			stereo *= 1.0 / f32(num_aa_samples);

			bg_color = mix(vec4f(0.0), mix(color1 * 0.4, color2 * 0.4, stereo), cvg);
		}
		case 12u: {
			/* derived from https://www.shadertoy.com/view/NlsXDH */
			var p = uv;
			let _mult = 1.0;
			p = abs(vec2(_mult) - mymod(p * vec2(1.0 / _mult), vec2(_mult * 2.0))) - _mult;

			var t = uniforms.scene_time;

			let num_iterations = clamp(1 + uniforms.val_seq_1.x, 1, 32);
			let shift    = f32(uniforms.val_seq_1.y) * (1.0 / f32(0xff));
			let rotation = 4.0 * 3.14159265 * f32(uniforms.val_seq_1.z) * (1.0 / f32(0xff)) * (1.0 / 16.0);
			let it_shift = f32(uniforms.val_seq_1.w) * (1.0 / f32(0xff));
			let density  = f32(uniforms.val_seq_2.x) * (1.0 / f32(0xff));

			var ot = 1000.0;
			for(var i = 0; i < num_iterations; i++) {
				p = abs(rotation_matrix_2d(rotation * t) * p) / clamp(p.x * p.y, 0.125, 2.0) - 1.0 + it_shift;
				ot = min(ot, abs(p.x) * (1.0 - density) + 0.7 * fract(abs(p.y) *.05 + t *.25 + f32(i) * shift));
			}
			ot = exp(-10.0 * ot);
			bg_color = vec4f(ot * ot * ot, ot, ot * ot, 1.0) * 0.5;
		}
		case 13u: {
			/* derived from https://www.shadertoy.com/view/MfSBzG */
			var col = vec3(0.0);

			let t = uniforms.scene_time;
			let zoom = t;
			let rotation = 0.0;

			col.x = uv_fx.x;
			col.y = uv_fx.y;

			let _uv = rotation_matrix_2d(rotation) * uv_fx;

			let color_shift = f32(uniforms.val_seq_1.x) * (1.0 / f32(0xff));

			let scale = 32.0 * f32(uniforms.val_seq_1.y) * (1.0 / f32(0xff));

			let off_scale = 0.01 + 32.0 * f32(uniforms.val_seq_1.z) * (1.0 / f32(0xff));

			for (var i = 0; i < 3; i++) {
				var uv = _uv;

				let radialOffset = vec2(sin(zoom + length(uv) * scale), cos(zoom + length(uv) * scale));
				uv += radialOffset * off_scale;
				let lum = 0.2 / (length(fract(uv * 1.0) - 0.5) + 0.1);

				// Modify this line to introduce color shifting
				col[i] = lum * abs(sin(zoom + length(uv) * 2.0 + f32(i) * color_shift));
			}

			bg_color = vec4(col * 0.5, 1.0);
		}
		default: {
		}
	}

	bg_color = vec4(bg_color.rgb * uniforms.track_volume, 1.0);

	return bg_color;
}


fn render_sine(uv: vec2f) -> vec3f
{
	let speed = 0.03;
	var a = sin(uv.x + uniforms.scene_time * 1.0    * speed) * 0.85;
	var b = sin(uv.x + uniforms.scene_time * 2.0    * speed) * 0.85;
	var c = sin(uv.x + uniforms.scene_time * 3.0    * speed) * 0.85;

	a += 0.2 * sin(uv.x * 4.0 + uniforms.scene_time * 2.0 * speed) * sin(uniforms.scene_time * speed);
	b += 0.2 * sin(uv.x * 4.0 + uniforms.scene_time * 3.0 * speed) * sin(uniforms.scene_time * speed);
	c += 0.2 * sin(uv.x * 4.0 + uniforms.scene_time * 4.0 * speed) * sin(uniforms.scene_time * speed);

	a += 0.06 * sin(uv.x * 16.0 + uniforms.scene_time * 9.0  * speed) * 0.0;
	b += 0.06 * sin(uv.x * 16.0 + uniforms.scene_time * 10.0 * speed) * 0.0;
	c += 0.06 * sin(uv.x * 16.0 + uniforms.scene_time * 11.0 * speed) * 0.0;

	let aa = 0.05 / abs(uv.y + a);
	let bb = 0.05 / abs(uv.y + b);
	let cc = 0.05 / abs(uv.y + c);

	let color_a = vec3f(1.0, 0.1, 0.1);
	let color_b = vec3f(0.1, 1.0, 0.1);
	let color_c = vec3f(0.1, 0.1, 1.0);

	let sum = tanh(aa * color_a + bb * color_b + cc * color_c);


	return sum;
}

fn render_tunnel(uv: vec2f) -> vec3f
{
	let angle = atan2(uv.y, uv.x);
	let r = length(uv);
	let st = vec2(angle / 3.14159, 0.1 / r) * 16.0 + sin(vec2(uniforms.scene_time, uniforms.scene_time) * 0.1) * 10.0;

	let seq_step = (uniforms.sequencer_ticks / 12u) % 16;
	let color_a = sin(st.x * 3.14159 * f32(uniforms.val_seq_1.x + 1) + st.y) * 0.5 + vec2(0.5, 0.5);
	let color_b = sin(-(st.x + f32(seq_step))* f32(uniforms.val_seq_1.y + 1) * 3.14159 / 4.0 + st.y) * 0.5 + vec2(0.5, 0.5);

	var color_fg = textureLoad(tex_colormap, vec2i(14, 0), 0).rgb * r;
	var color_bg = textureLoad(tex_colormap, vec2i(1,  0), 0).rgb * r * 0.75;

	return color_a.x * color_fg + color_b.x * color_bg;
}

fn rotation_matrix_2d(angle: f32) -> mat2x2<f32>
{
	let c = cos(angle);
	let s = sin(angle);
	return mat2x2<f32>(c, -s, s, c);
}

/* https://www.youtube.com/watch?v=cQXAbndD5CQ */
fn render_xor_circles(_uv: vec2f) -> vec3f
{
	var sum = vec3f(0.0);

	let lookup_radius = 1;
	var accum = 0.0;

	if true {
		let rot_speed = 0.05;
		let mult = 1.0 + f32(uniforms.val_seq_1.x);
		let uv_rot = rotation_matrix_2d(uniforms.scene_time * rot_speed) * (_uv * mult);
		let id = floor(uv_rot);
		let uv = fract(uv_rot) - 0.5;
		for(var yy = -lookup_radius; yy <= lookup_radius; yy++) {
			for(var xx = -lookup_radius; xx <= lookup_radius; xx++) {
				let off = vec2f(f32(xx), f32(yy));
				let t = f32(uniforms.sequencer_ticks % (64 * 6)) * 3.141592653589 / (32.0 * 6.0) + length(id + off) * 5.0;

				let r = mix(0.2, 1.5, sin(t) * 0.5 + 0.5);

				let dist = length(uv - off);
				let f = smoothstep(r, r * 0.92, dist);

				accum = (1.0 - f) * accum + (1.0 - accum) * f;
			}
		}
	}

	if uniforms.val_seq_1.y > 0 {
		let rot_speed = -0.0213123132;
		let mult = f32(uniforms.val_seq_1.y);
		let uv_rot = rotation_matrix_2d(uniforms.scene_time * rot_speed) * (_uv * mult);
		let id = floor(uv_rot);
		let uv = fract(uv_rot) - 0.5;
		for(var yy = -lookup_radius; yy <= lookup_radius; yy++) {
			for(var xx = -lookup_radius; xx <= lookup_radius; xx++) {
				let off = vec2f(f32(xx), f32(yy));
				let t = f32(uniforms.sequencer_ticks % (64 * 6)) * 3.141592653589 / (16.0 * 6.0) + length(id + off) * 8.0;

				let r = mix(0.2, 1.5, sin(t) * 0.5 + 0.5);

				let dist = length(uv - off);
				let f = smoothstep(r, r * 0.92, dist);

				accum = (1.0 - f) * accum + (1.0 - accum) * f;
			}
		}
	}


	sum += accum * 0.1 * textureLoad(tex_colormap, vec2i(1, 0), 0).rgb;

	return sum;
}


fn sd_octahedron(pos: vec3f, s: f32) -> f32
{
	let p = abs(pos);
	return (p.x + p.y + p.z - s) * 0.57735027;
}

fn map(pos: vec3f) -> f32
{
	let p_repeat = fract(pos) - 0.5;
	return sd_octahedron(p_repeat, 0.25);
}

fn rot_2d(angle: f32) -> mat2x2<f32>
{
	let s = sin(angle);
	let c = cos(angle);
	return mat2x2<f32>(c, -s, s, c);
}


fn render_distance_field(uv: vec2f) -> vec3f
{
	let r_o = vec3f(0.0, 0.0, -uniforms.scene_time * 0.2);
	let r_d = normalize(vec3f(uv.x, -uv.y, -2.0));
	

	var t = 0.0;

	var i = 0;
	let max_steps = 128;
	var pos = vec3f(0.0, 0.0, 0.0);
	for(; i < max_steps; i++) {
		pos = r_o + r_d * t + vec3f(sin(t * 1.3) * 0.3, sin(t) * 0.2, 0.0);

		let pos_2d = pos.xy * rot_2d(pos.z * 0.1);

		pos.x = pos_2d.x;
		pos.y = pos_2d.y;

		let d = map(pos);

		if d < 0.001 || t > 100.0 {
			break;
		}

		t += d;
	}

	let it = f32(i) / f32(max_steps);

	let p_cell = vec3i(floor(pos));

	let seq_step = (uniforms.sequencer_ticks / 6u) % 64u;

	let fog = exp(-t * 0.08);

	let color = select(vec3f(0.1, 0.1, 0.1), vec3f(0.0, 1.0, 1.0),
	(p_cell.z + 15 * p_cell.x + 21 * p_cell.y) %
	16 == i32(seq_step) % 16);

	return color * it * fog;
}

fn hash33(p3_: vec3f) -> vec3f
{
	var p3 = fract(p3_ * vec3f(0.1031, 0.11369, 0.13787));
    p3 += dot(p3, p3.yxz + 19.19);
    return -1.0 + 2.0 * fract(vec3f((p3.x + p3.y)*p3.z, (p3.x+p3.z)*p3.y, (p3.y+p3.z)*p3.x));
}

fn noise(p: vec3f) -> f32
{
    let K1 = 0.333333333;
    let K2 = 0.166666667;
    
    let i = floor(p + (p.x + p.y + p.z) * K1);
    let d0 = p - (i - (i.x + i.y + i.z) * K2);
    
    // thx nikita: https://www.shadertoy.com/view/XsX3zB
    let e = step(vec3f(0.0), d0 - d0.yzx);
	let i1 = e * (1.0 - e.zxy);
	let i2 = 1.0 - e.zxy * (1.0 - e);
    
    let d1 = d0 - (i1 - 1.0 * K2);
    let d2 = d0 - (i2 - 2.0 * K2);
    let d3 = d0 - (1.0 - 3.0 * K2);
    
    let h = max(vec4f(0.6) - vec4f(dot(d0, d0), dot(d1, d1), dot(d2, d2), dot(d3, d3)), vec4f(0.0));
    let n = h * h * h * h * vec4f(dot(d0, hash33(i)), dot(d1, hash33(i + i1)), dot(d2, hash33(i + i2)), dot(d3, hash33(i + 1.0)));
    
    return dot(vec4f(31.316), n);
}

fn noise_sum_abs(p_: vec3f) -> f32
{
    var f = 0.0;
    var p = p_ * 2.0;
    f += 1.0000 * abs(noise(p)); p = 2.0 * p;
    f += 0.5000 * abs(noise(p)); p = 2.0 * p;
	f += 0.2500 * abs(noise(p)); p = 2.0 * p;
	f += 0.1250 * abs(noise(p)); p = 2.0 * p;
	f += 0.0625 * abs(noise(p)); p = 2.0 * p;
    
    return f;
}

fn noise_sum_abs_sin(p: vec3f) -> vec3f
{
    var f = noise_sum_abs(p);
    f = sin(uniforms.elapsed_time * 0.25 + f * 2.5 + p.x * 5.0 + p.y * p.y * 5.0);

	let color_fg = textureLoad(tex_colormap, vec2i(1, 0), 0).rgb;
	let color_bg = textureLoad(tex_colormap, vec2i(0, 0), 0).rgb;

	//let color_fg = vec3(0.1, 1.0, 1.0);
	//let color_bg = vec3(0.0, 0.0, 0.0);

	return tanh(vec3(color_fg * 0.1 * pow(max(1e-3, 1.0 - sin(f * 4.0) * 0.4999 - 0.5), -0.8)));
    
    //return mix(color_fg, color_bg, sin(f) * 0.5 + 0.5);
    //return pow(mix(color_fg, color_bg, sin(f) * 0.5 + 0.5), vec3f(8.2));
}

@compute @workgroup_size(16, 16, 1)
fn populate_frame_buffer(@builtin(global_invocation_id) _id: vec3u)
{
	var id = _id;

	var shift = 1u << u32(uniforms.screen_flash * 4.0 + 1);

	if(uniforms.screen_flash > 0.0) {
		id.x ^= shift;
		id.y ^= shift << 2;
		id.y ^= shift << 1;

		//id.x ^= (shift >> 4u);
		id.y ^= (shift >> 4u);
	}

	let atlas_size  = textureDimensions(tex_charset);
	let char_size   = vec2i(i32(atlas_size.x) / 16, i32(atlas_size.y) / 16);
	var char_info   = textureLoad(tex_text_buffer, vec2i(id.xy) / char_size, 0);
	let char_offset = vec2i(id.xy) % char_size;

	let glyph_pos   = vec2i(i32(char_info.x) % 16, i32(char_info.x) / 16) * char_size + char_offset;

	var color_fg = textureLoad(tex_colormap, vec2i(i32(char_info.y), 0), 0).rgb;
	var color_bg = textureLoad(tex_colormap, vec2i(i32(char_info.z), 0), 0).rgb;

	let c = textureLoad(tex_charset, glyph_pos, 0).r;

	if(char_info.w == 1u || char_info.w == 4u) {
		color_fg = textureLoad(tex_colormap, vec2i(14, 0), 0).rgb;
		color_fg *= 0.6 + 0.4 * sin(uniforms.elapsed_time * 3.14159 * 2.0 * 2.0);
	}
	if(char_info.w == 3u) {
		color_bg *= 0.8 + 0.2 * sin(uniforms.elapsed_time * 3.14159 * 2.0 * 1.0);
	}

	var color = mix(color_bg, color_fg, c);

	if(char_info.w == 4u) { // hack to make playback cursor look pretty in sequencer page
		let blend = 0.5 + 0.5 * sin(uniforms.elapsed_time * 3.14159 * 2.0 * 2.0);

		
		let glyph_pos_col = vec2i(i32(0xb3) % 16, i32(0xb3) / 16) * char_size + char_offset;
		var c_col = textureLoad(tex_charset, glyph_pos_col, 0).r;

		c_col -= c;

		color += max(0.0, c_col) * textureLoad(tex_colormap, vec2i(i32(char_info.y), 0), 0).rgb;

		char_info.w = 1u;
	}


	let res = vec2f(120.0 * f32(char_size.x), 68.0 * f32(char_size.y));
	let aspect = vec2f(res.y / res.x, 2.0);


	if false {
		let uv = vec2(f32(id.x), f32(id.y)) / res;
		let bg_effect = noise_sum_abs_sin(vec3f(uv, uniforms.elapsed_time * 0.01));
		color += 0.1 * bg_effect;

		color = mix(color, bg_effect, 1.0);
	}

	if false {
		let uv = aspect * ((vec2(f32(id.x), f32(id.y)) / res) * 2.0 - vec2(1.0, 1.0));
		color += render_sine(uv) * 0.1;
	}

	//color = mix(color, vec3(1.0) - color, uniforms.screen_flash * 0.25);

	textureStore(tex_storage_pong, vec2i(_id.xy), vec4(color, f32(char_info.w)));
}

@compute @workgroup_size(16, 16, 1)
fn blur_horizontally(@builtin(global_invocation_id) id: vec3u)
{
	var sum = vec3(0.0, 0.0, 0.0);
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(-2, 0), 0).rgb * filter_weights[2];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(-1, 0), 0).rgb * filter_weights[1];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i( 0, 0), 0).rgb * filter_weights[0];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i( 1, 0), 0).rgb * filter_weights[1];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i( 2, 0), 0).rgb * filter_weights[2];

	textureStore(tex_storage_pong, vec2i(id.xy), vec4(sum, 1.0));
}

@compute @workgroup_size(16, 16, 1)
fn blur_vertically(@builtin(global_invocation_id) id: vec3u)
{
	var sum = vec3(0.0, 0.0, 0.0);
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(0, -2), 0).rgb * filter_weights[2];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(0, -1), 0).rgb * filter_weights[1];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(0,  0), 0).rgb * filter_weights[0];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(0,  1), 0).rgb * filter_weights[1];
	sum += textureLoad(tex_ping, vec2i(id.xy) + vec2i(0,  2), 0).rgb * filter_weights[2];

	textureStore(tex_storage_pong, vec2i(id.xy), vec4(sum, 1.0));
}
